# version avec => au lieu de <=> dans les points fixes

import os
#import transformation

# On a ici les classes pour l'AST d'une formule.

# Les classes de formules servent a la fois a manipuler des formules QCTL,
# et a manipuler les formules de QBF qu'on soumettra a Z3,
# sachant que QBF est inclus dans QCTL.

# La methode toString d'une formule permet ainsi d'obtenir,
# si cette formule est dans QBF, sa representation en forme prefixe
# particuliere a Z3.

# On donne aussi une methode toString aux formules qui ne sont pas dans QBF
# pour faciliter le debug.

# La methode write ecrit la representation d'une formule QBF dans un fichier.
# Elle est plus efficace que toString, qui fait des additions (couteuses) de string.

# la methode nbalt renvoie le nb d'alternation E/A ou A/E. La var t sert a indiquer si on est dans un E (0) ou dans A (1), ou dans du Boolean (None)

# PropQuant renvoie la liste des var introduites par Exists/Forall et celle introduite par Exists1/Forall1
# (pas le meme traitement dans transformation 5)

# La methode MaPl (mise a plat) renvoie une formule de la forme suivante si niv==-1 :
# [Q] Exists q1...qk . Phi1 ^ AG (q1 <=> def1) ^... ^AG (qk <=> defk)
# ou Phi1 et les def_i sont de heuteur temporelle 1 [les def_i ont meme un op temporel a la racine]
# et [Q] est le bloc de quantificateurs initial (forme normale prenex).
# si on l'appelle avec niv=0, on obtient:
# [Q] Phi1
# dans ce dernier cas, les def_i sont a recuperer dans les parametres D et Df.
# Et c'est ce qu'on utilise pour coder vers z3...
# On utilise un booleen justeX pour ne garder dans Phi1 que une hauteur de EX/AX de 1, et
# pas de U: c'est pour homogeneiser le codage z3 


# La methode AGFlat2QBF est utilisee pour coder les ssf AG(p <=> def) ou def est une formule CTL de HT <=1.
# Elle renvoie une liste de termes (conjonctions) representant la reduction.
# La methode FFlat2z3 sert a` coder une formule depuis un etat s. Elle est utilisee depuis AGFlat2QBF
# et depuis transformation5 pour la formule CTL "principale" (sans les AG(p <=> def) )

count=0

def getIdent() :
    global count
    count+=1
    return "q_"+str(count)

def Kripkestring_of_state(st) :
        ans=st[0]
        for c in st[1:] :
            ans=ans+"_"+c
        return ans
    

###############################################################################
#                          AP
###############################################################################
class AP :
    def __init__(self, val) :
        self.fils=[]
        self.ap=val

    def toString(self) :
        return self.ap

    def SimplifB(self) :
        return AP(self.ap)

    def PropQuant(self) : 
        return ([],[],[])
    
    def write(self, output) :
        output.write(self.ap)

    def nbalt(self,t = None) :
        return 0

    def isPNF(self,niveau) :  # niveau = True si des quantif sont encore autorisees
        return True

    def MaPl(self,D,Df,niv,justeX =False) :   # retourne la formule mise a plat (de HT 1). D associe a une sous-form une chaine (le nom de la prop), niv=-1 sert a ajouter des Exists pour les formules de CTL, niv=0 (resp. 1) sert a indiquer la HT de la formule. 
        return AP(self.ap)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,AP(self.ap))
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        (PropB, PropBV, PropQ1, PropQ1FA) = VarQ
        if self.ap in PropB :
            fout.write(" "+self.ap+"_"+Kripkestring_of_state(stp)+" ")
        elif self.ap in PropBV :
            fout.write(" (bvult "+self.ap+"_"+Kripkestring_of_state(stp)+" "+krip.cBUntil()+")")
        elif self.ap in PropQ1 :
            fout.write(" (= "+self.ap+" "+krip.cNumero(stp)+")")
        elif self.ap in krip.label[stp] :
            fout.write(" true ")
        else :
            fout.write(" false ")
        
    def FNN(self,np) : # mise sous forme normale negative, np=True ssi on propage un non
        if np :
            return NOT(AP(self.ap))
        else :
            return AP(self.ap)

###############################################################################
#                          QUANT
###############################################################################
# Une proposition atomique quantifiee
class QUANT(AP) :
    pass

###############################################################################
#                          NOT
###############################################################################
class NOT :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(not "+self.fils[0].toString()+")"

    def SimplifB(self) :
        return (NOT(self.fils[0].SimplifB()))

    def PropQuant(self) :
        return self.fils[0].PropQuant()
    
    def write(self, output) :
        output.write("(not ")
        self.fils[0].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        return self.fils[0].nbalt(t)

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(niveau)
    
    def MaPl(self,D,Df,niv,justeX =False) :
        aux = self.fils[0]
        if not isinstance(aux,AP) : 
            print("Warning ! Dans Mise-a-plat : formule non mise en FNN ! (NOT)")
        return (NOT(self.fils[0].MaPl(D,Df,niv,justeX)))

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write("(not ")
        self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        fout.write(")")
        
    def FNN(self,np) : 
        return self.fils[0].FNN(not np)

###############################################################################
#                          AND
###############################################################################
class AND :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        return "(and "+self.fils[0].toString()+" "+self.fils[1].toString()+")"

    def SimplifB(self) :
        lres=[]
        for f in self.fils :
            fs=f.SimplifB()
            if isinstance(fs,TOP) :
                pass
            elif isinstance(fs,BOTTOM) :
                return BOTTOM()
            elif isinstance(fs,AND) or isinstance(fs,SEQAND) :
                lres.extend(fs.fils)
            else :
                lres.append(fs)
        if len(lres)==0 :
            return TOP()
        elif len(lres)==1 :
            return lres[0]
        else :
            return SEQAND(lres)

    def PropQuant(self) :
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)        
        return (L,L1,L2)
    
    def write(self, output) :
        output.write("(and ")
        self.fils[0].write(output)
        output.write(" ")
        self.fils[1].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))

    def isPNF(self,niveau) :
        for f in self.fils :
            if not f.isPNF(niveau) :
                return False
        return True

    def MaPl(self,D,Df,niv,justeX =False) :
        lres=[]
        for f in self.fils :
            lres.append(f.MaPl(D,Df,max(0,niv),justeX))
        if niv==-1 : # on doit ajouter le bloc Exists
            return AuxMaPl(AND(lres[0],lres[1]),D,Df)
        return AND(lres[0],lres[1])

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write("(and ")
        self.fils[0].FFlat2z3(VarQ,krip,stp,fout)
        fout.write(" ")
        self.fils[1].FFlat2z3(VarQ,krip,stp,fout)
        fout.write(")")

    def FNN(self,np) :
        lres=[]
        for f in self.fils :
            lres.append(f.FNN(np))
        if np :
            return OR(lres[0],lres[1])
        else :
            return AND(lres[0],lres[1])

###############################################################################
#                          OR
###############################################################################
class OR :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]
        
    def toString(self) :
        return "(or "+self.fils[0].toString()+" "+self.fils[1].toString()+")"

    def SimplifB(self) :
        lres=[]
        for f in self.fils :
            fs=f.SimplifB()
            if isinstance(fs,BOTTOM) :
                pass
            elif isinstance(fs,TOP) :
                return TOP()
            elif isinstance(fs,OR) or isinstance(fs,SEQOR) :
                lres.extend(fs.fils)
            else :
                lres.append(fs)
        if len(lres)==0 :
            return BOTTOM()
        elif len(lres)==1 :
            return lres[0]
        else :
            return SEQOR(lres)

    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)        
        return (L,L1,L2)
    
    def write(self, output) :
        output.write("(or ")
        self.fils[0].write(output)
        output.write(" ")
        self.fils[1].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))

    def isPNF(self,niveau) :
        for f in self.fils :
            if not f.isPNF(niveau) :
                return False
        return True
    
    def MaPl(self,D,Df,niv,justeX =False) :
        lres=[]
        for f in self.fils :
            lres.append(f.MaPl(D,Df,max(0,niv),justeX))
        if niv==-1 : # on doit ajouter le bloc Exists
            return AuxMaPl(OR(lres[0],lres[1]),D,Df)
        return OR(lres[0],lres[1])

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write("(or ")
        self.fils[0].FFlat2z3(VarQ,krip,stp,fout)
        fout.write(" ")
        self.fils[1].FFlat2z3(VarQ,krip,stp,fout)
        fout.write(")")

    def FNN(self,np) :
        lres=[]
        for f in self.fils :
            lres.append(f.FNN(np))
        if np :
            return AND(lres[0],lres[1])
        else :
            return OR(lres[0],lres[1])

###############################################################################
#                          SEQAND
###############################################################################
# Le 'et' et 'ou' de z3 prennent un nombre variable d'arguments.
# On definit donc un ET et un OU qui on un nombre variable de fils.
class SEQAND :
    def __init__(self, fList) :
        self.fils=fList

    def toString(self) :
        ans=""
        for f in self.fils :
            ans=ans+" "+f.toString()
        return "(and "+ans+")"

    def SimplifB(self) :
        lres=[]
        for f in self.fils :
            fs=f.SimplifB()
            if isinstance(fs,TOP) :
                pass
            elif isinstance(fs,BOTTOM) :
                return BOTTOM()
            elif isinstance(fs,AND) or isinstance(fs,SEQAND) :
                lres.extend(fs.fils)
            else :
                lres.append(fs)
        if len(lres)==0 :
            return TOP()
        elif len(lres)==1 :
            return lres[0]
        else :
            return SEQAND(lres)

    def PropQuant(self) : 
        lres,lres1,lres2=[],[],[]
        for f in self.fils :
            (L,L1,L2) = f.PropQuant()
            lres.extend(L)
            lres1.extend(L1)
            lres2.extend(L2)
        return (lres,lres1,lres2)

    
    def write(self, output) :
        output.write("(and")
        for f in self.fils :
            output.write(" ")
            f.write(output)
        output.write(")")

    def nbalt(self,t = None) :
        res = 0
        for f in self.fils :
            res = max(res,f.nbalt(t))
        return res

    def isPNF(self,niveau) :
        for f in self.fils :
            if not f.isPNF(niveau) :
                return False
        return True

    def MaPl(self,D,Df,niv,justeX =False) :
        lres=[]
        for f in self.fils :
            lres.append(f.MaPl(D,Df,max(0,niv),justeX))
        if niv==-1 : # on doit ajouter le bloc Exists
            return AuxMaPl(SEQAND(lres),D,Df)
        return SEQAND(lres)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write("(and ")
        for f in self.fils :
            f.FFlat2z3(VarQ,krip,stp,fout,D)
            fout.write(" ")
        fout.write(")")
        
    def FNN(self,np) :
        lres=[]
        for f in self.fils :
            lres.append(f.FNN(np))
        if np :
            return SEQOR(lres)
        else :
            return SEQAND(lres)

###############################################################################
#                          SEQOR
###############################################################################
class SEQOR :
    def __init__(self, fList) :
        self.fils=fList

    def toString(self) :
        ans=""
        for f in self.fils :
            ans=ans+" "+f.toString()
        return "(or "+ans+")"
    
    def SimplifB(self) :
        lres=[]
        for f in self.fils :
            fs=f.SimplifB()
            if isinstance(fs,BOTTOM) :
                pass
            elif isinstance(fs,TOP) :
                return TOP()
            elif isinstance(fs,OR) or isinstance(fs,SEQOR) :
                lres.extend(fs.fils)
            else :
                lres.append(fs)
        if len(lres)==0 :
            return BOTTOM()
        elif len(lres)==1 :
            return lres[0]
        else :
            return SEQOR(lres)

    def PropQuant(self) : 
        lres,lres1,lres2=[],[],[]
        for f in self.fils :
            (L,L1,L2) = f.PropQuant()
            lres.extend(L)
            lres1.extend(L1)
            lres2.extend(L2)
        return (lres,lres1,lres2)
    
    def write(self, output) :
        output.write("(or")
        for f in self.fils :
            output.write(" ")
            f.write(output)
        output.write(")")

    def nbalt(self,t = None) :
        res = 0
        for f in self.fils :
            res = max(res,f.nbalt(t))
        return res
    
    def isPNF(self,niveau) :
        for f in self.fils :
            if not f.isPNF(niveau) :
                return False
        return True
    
    def MaPl(self,D,Df,niv,justeX =False) :
        lres=[]
        for f in self.fils :
            lres.append(f.MaPl(D,Df,max(0,niv),justeX))
        if niv==-1 : # on doit ajouter le bloc Exists
            return AuxMaPl(SEQOR(lres),D,Df)
        return SEQOR(lres)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write("(or ")
        for f in self.fils :
            f.FFlat2z3(VarQ,krip,stp,fout,D)
            fout.write(" ")
        fout.write(")")
    
    def FNN(self,np) :
        lres=[]
        for f in self.fils :
            lres.append(f.FNN(np))
        if np :
            return SEQAND(lres)
        else :
            return SEQOR(lres)
    
    
def fusionAND(f1, f2) :
    return SEQAND(f1.fils+f2.fils)

def fusionOR(f1, f2) :
    return SEQOR(f1.fils+f2.fils)

###############################################################################
#                          EX
###############################################################################
class EX :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(EX "+self.fils[0].toString()+")"

    def SimplifB(self) :
        fs=self.fils[0].SimplifB()
        if isinstance(fs,BOTTOM) :
            return BOTTOM()
        else :
            return EX(fs)

    def PropQuant(self) : 
        return self.fils[0].PropQuant()
 
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return self.fils[0].nbalt(t)

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False)

    def MaPl(self,D,Df,niv,justeX =False) :
        # si niv <1, on descend...
        # sinon on cherche si self est deja dans D, et si oui on renvoie la prop associee
        # sinon on renvoie une nouvelle prop que l'on insere dans D
        f1 = self.fils[0].MaPl(D,Df,1,False)
        if niv==-1 :
            return AuxMaPl(EX(f1),D,Df)
        elif niv==0 : 
            return EX(f1)
        else : 
            aux=EX(f1)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(= ")
            fp.FFlat2z3(VarQ,krip,st,fout,D)
            if len(krip.succ[st])==1 :
                fout.write(" ")
                for stp in krip.succ[st] : 
                    f1.FFlat2z3(VarQ,krip,stp,fout,D)
            else :
                fout.write("(or ")
                for stp in krip.succ[st] :# pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
                    f1.FFlat2z3(VarQ,krip,stp,fout,D) # renvoie la formule Booleenne f1 "sur" st' !
                    fout.write(" ")
                fout.write(")")
            fout.write(")\n")
        

    def FFlat2z3(self,VarQ,krip,st,fout,D)  :
        f1=self.fils[0]
        if len(krip.succ[st])==1 :
                for stp in krip.succ[st] : 
                        f1.FFlat2z3(VarQ,krip,stp,fout,D)
        else :
            fout.write("(or ")
            for stp in krip.succ[st] :# on doit avoir:  \/_{st->st'} p_f1_st' 
                f1.FFlat2z3(VarQ,krip,stp,fout,D) # pas optimal ! 
            fout.write(")")
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return AX(f)
        else :
            return EX(f)

###############################################################################
#                          AX
###############################################################################
class AX :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(AX "+self.fils[0].toString()+")"

    def SimplifB(self) :
        fs=self.fils[0].SimplifB()
        if isinstance(fs,TOP) :
            return TOP()
        else :
            return AX(fs)

    def PropQuant(self) : 
        return self.fils[0].PropQuant()
 
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
#        return 0
        return self.fils[0].nbalt(t)


    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False)

    def MaPl(self,D,Df,niv,justeX =False) :
        f1 = self.fils[0].MaPl(D,Df,1,False)
        if niv == -1 :
                return AuxMaPl(AX(f1),D,Df)
        elif niv==0 : 
            return AX(f1)
        else : 
            aux=AX(f1) 
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(= ")
            fp.FFlat2z3(VarQ,krip,st,fout,D)
            if len(krip.succ[st])==1 :
                fout.write(" ")
                for stp in krip.succ[st] :
                    f1.FFlat2z3(VarQ,krip,stp,fout,D)  
            else :
                fout.write("(and ")
                for stp in krip.succ[st] :# pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
                    f1.FFlat2z3(VarQ,krip,stp,fout,D) # renvoie la formule Booleenne f1 "sur" st' !
                    fout.write(" ")
                fout.write(")")
            fout.write(")\n")

    def FFlat2z3(self,VarQ,krip,st,fout,D)  :
        f1=self.fils[0]
        if len(krip.succ[st])==1 :
                for stp in krip.succ[st] : 
                        f1.FFlat2z3(VarQ,krip,stp,fout,D)
        else :
            fout.write("(and ")
            for stp in krip.succ[st] :# on doit avoir:  /\_{st->st'} p_f1_st' 
                f1.FFlat2z3(VarQ,krip,stp,fout,D) # pas optimal ! 
            fout.write(")")
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return EX(f)
        else :
            return AX(f)

###############################################################################
#                          EU
###############################################################################
class EU :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        return "(E "+self.fils[0].toString()+" U "+self.fils[1].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        f2=self.fils[1].SimplifB()
        if isinstance(f2,TOP) :
            return TOP()
        elif isinstance(f1,BOTTOM) :
            return f2
        else :
            return EU(f1,f2)  # on pourrait renvoyer EF si f1=T...

    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)
        return (L,L1,L2)
    
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
#        return 0
        return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False) and self.fils[1].isPNF(False)

    def MaPl(self,D,Df,niv,justeX =False) :
        aux1 = self.fils[0].MaPl(D,Df,1,False)
        aux2 = self.fils[1].MaPl(D,Df,1,False)
        if niv == -1 :
            return AuxMaPl(EU(aux1,aux2),D,Df)
        elif niv == 0 and justeX==False: 
            return EU(aux1,aux2)
        else : 
            aux=EU(aux1,aux2)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
#            if niv == -1 :
#                return AuxMaPl(AP(prop),D,Df)
            return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        f2=self.fils[1]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(=> ")      # => vs <=>
            fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(0)+") ")
            f2.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(") ")
            for d in range(1,krip.BUntil) :
                fout.write("(=> ")   # => vs <=>
                fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(d)+") ")
                fout.write(" (and ")
                f1.FFlat2z3(VarQ,krip,st,fout,D)
                fout.write(" ")
                if len(krip.succ[st])==1 :
                    for stp in krip.succ[st] : 
                        fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d-1)+") ")
                else :
                    fout.write("(or ")
                    for stp in krip.succ[st] :
                        fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d-1)+") ")
                        fout.write(" ")
                    fout.write(")")
                fout.write("))")
        fout.write("\n")
        
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        
        
    def FNN(self,np) :
        f1=self.fils[0].FNN(np)
        f2=self.fils[1].FNN(np)
        if np :
            return AW(f2,AND(f1,f2))
        else :
            return EU(f1,f2)
            
###############################################################################
#                          AU
###############################################################################
class AU :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        return "(A "+self.fils[0].toString()+" U "+self.fils[1].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        f2=self.fils[1].SimplifB()
        if isinstance(f2,TOP) :
            return TOP()
        elif isinstance(f1,BOTTOM) :
            return f2
        else :
            return AU(f1,f2)  # on pourrait renvoyer AF si f1=T...

    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)
        return (L,L1,L2)
    
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
#        return 0
        return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False) and self.fils[1].isPNF(False)

    def MaPl(self,D,Df,niv,justeX =False) :
        aux1 = self.fils[0].MaPl(D,Df,1,False)
        aux2 = self.fils[1].MaPl(D,Df,1,False)
        if niv == -1 :
            return AuxMaPl(AU(aux1,aux2),D,Df)
        elif niv == 0 and justeX==False: 
            return AU(aux1,aux2)
        else : 
            aux=AU(aux1,aux2)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
#            if niv == -1 :
#                return AuxMaPl(AP(prop),D,Df)
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        f2=self.fils[1]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(=> ")   # => vs <=>
            fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(0)+") ")
            f2.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(") ")
            for d in range(1,krip.BUntilD()) :
                fout.write("(=> ")    # => vs <=>
                fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(d)+") ")
                fout.write(" (and ")
                f1.FFlat2z3(VarQ,krip,st,fout,D)
                fout.write(" ")
                for stp in krip.succ[st] :
                    fout.write(" (bvult "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d)+") ")
                    fout.write(" ")
                fout.write(")")
            fout.write(")")
        fout.write("\n")
        
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        
    def FNN(self,np) :
        f1=self.fils[0].FNN(np)
        f2=self.fils[1].FNN(np)
        if np :
            return EW(f2,AND(f1,f2))
        else :
            return AU(f1,f2)

###############################################################################
#                          EW
###############################################################################
class EW :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        return "(E "+self.fils[0].toString()+" W "+self.fils[1].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        f2=self.fils[1].SimplifB()
        if isinstance(f1,TOP) or  isinstance(f2,TOP) :
            return TOP()
        elif isinstance(f1,BOTTOM) :
            return f2
        else :
            return EW(f1,f2)  # on pourrait renvoyer EG si f2=bot...
        
    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)
        return (L,L1,L2)
    
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
#        return 0
        return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False) and self.fils[1].isPNF(False)

    def MaPl(self,D,Df,niv,justeX =False) :
        aux1 = self.fils[0].MaPl(D,Df,1,False)
        aux2 = self.fils[1].MaPl(D,Df,1,False)
        if niv == -1  : 
            return AuxMaPl(EW(aux1,aux2),D,Df)
        elif niv == 0 and justeX==False: 
            return EW(aux1,aux2)
        else : 
            aux=EW(aux1,aux2)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        f2=self.fils[1]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(=> ")   # => vs <=>
            fp.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write("(or ")
            f2.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" (and ")
            f1.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" ")
            if len(krip.succ[st])==1 :
                for stp in krip.succ[st] :
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
            else :
                fout.write("(or ")
                for stp in krip.succ[st] :
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
                    fout.write(" ")
                fout.write(")")
            fout.write(")))")
        fout.write("\n")
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        
    def FNN(self,np) :
        f1=self.fils[0].FNN(np)
        f2=self.fils[1].FNN(np)
        if np :
            return AU(f2,AND(f1,f2))
        else :
            return EW(f1,f2)

###############################################################################
#                          AW
###############################################################################
class AW :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        print(self.fils[0])
        print(self.fils[1])
        return "(A "+self.fils[0].toString()+" W "+self.fils[1].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        f2=self.fils[1].SimplifB()
        if isinstance(f1,TOP) or  isinstance(f2,TOP) :
            return TOP()
        elif isinstance(f1,BOTTOM) :
            return f2
        else :
            return AW(f1,f2)  # on pourrait renvoyer AG si f2=bot...

    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)
        return (L,L1,L2)
    
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
#        return 0
        return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False) and self.fils[1].isPNF(False)
    
    def MaPl(self,D,Df,niv,justeX = False) :
        aux1 = self.fils[0].MaPl(D,Df,1,justeX)
        aux2 = self.fils[1].MaPl(D,Df,1,justeX)
        if niv == -1 :
            return AuxMaPl(AW(aux1,aux2),D,Df)
        elif niv == 0 and justeX==False: 
            return AW(aux1,aux2)
        else : 
            aux=AW(aux1,aux2)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        f2=self.fils[1]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(=> ")   # => vs <=>
            fp.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write("(or ")
            f2.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" (and ")
            f1.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" ")
            if len(krip.succ[st])==1 :
                for stp in krip.succ[st] : 
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
            else :
                for stp in krip.succ[st] :
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
                    fout.write(" ")
            fout.write(")))")
        fout.write("\n")
        
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        
    def FNN(self,np) :
        f1=self.fils[0].FNN(np)
        f2=self.fils[1].FNN(np)
        if np :
            return EU(f2,AND(f1,f2))
        else :
            return AW(f1,f2)

###############################################################################
#                          FORALL
###############################################################################
#on peut mettre plusieurs variables sur le meme
#quantificateur en z3.
class FORALL :
    def __init__(self, varList, a) :
        self.fils=[a]
        self.var=varList

    def toString(self) :
        ans=""
        for v in self.var :
            ans=ans+"("+v+" Bool)"
        return "(forall ("+ans+") "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        if isinstance(f1,FORALL) :
            f1.var.extend(self.var)
            return f1
        return FORALL(self.var,f1)

    def PropQuant(self) :
        (lres,lres1,lres2) = self.fils[0].PropQuant()
        L=self.var[:]
        lres.extend(L)
        return (lres,lres1,lres2)
    
    def write(self, output) :
        output.write("(forall (")
        for v in self.var :
            output.write("(")
            output.write(v)
            output.write(" Bool)")
        output.write(") ")
        self.fils[0].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        if t==None :
            return self.fils[0].nbalt(1)
        elif t==0 :
            return 1+self.fils[0].nbalt(1)
        else :
            return self.fils[0].nbalt(1)

    def isPNF(self,niveau) :
        if not niveau :
            return False
        return self.fils[0].isPNF(niveau)
    
    def MaPl(self,D,Df,niv,justeX = False) :
        if (len(D) > 0) :
            print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (FORALL) !")
        # if niv != -1 :
        #     print("Warning ! Mauvaise valeur pour init dans la mise a plat d'une formule (FORALL) !")
        return FORALL(self.var,self.fils[0].MaPl(D,Df,niv,justeX))
        # sf=self.fils[0]
        # if isinstance(sf,FORALL) or isinstance(sf,FORALL1) or isinstance(sf,EXISTS) or isinstance(sf,EXISTS) :
        #     aux = self.fils[0].MaPl(D,Df)
        # else :
        #     aux = sf.MaPl(D,Df)
        #     aux = AuxMaPl(aux,D,Df)
        # return FORALL(self.var,aux)

    def AjoutPropQuant(self,L) :
        return FORALL(self.var,self.fils[0].AjoutPropQuant(L))
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        if (not stp == krip.start) :
            print("Warning ! FFlat2z3 - FORALL : pb d'arguments !")
        if len(self.var)==0 :
            self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        (PropB, PropBV, PropQ1, PropQ1FA) = VarQ
        fout.write("(forall (")
        for v in self.var :
            if v in PropB :
                for s in krip.accessibles[stp] : 
                    fout.write("("+v+"_"+Kripkestring_of_state(s)+" Bool)")
            elif v in PropBV :
                for s in krip.accessibles[stp] : 
                    fout.write("("+v+"_"+Kripkestring_of_state(s)+" (_ BitVec "+str(krip.TBVU)+"))")
            else :
                print("Cas bizarre dans FFlat2z3 - FORALL !")
        fout.write(")")
        self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        fout.write(")")
        
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return EXISTS(self.var,f)
        else :
            return FORALL(self.var,f)

###############################################################################
#                          FORALL1
###############################################################################
class FORALL1 :
    def __init__(self, varList, a) :
        self.fils=[a]
        self.var=varList

    def toString(self) :
        ans=""
        for v in self.var :
            ans=ans+"("+v+" Bool)"
        return "(forall1 ("+ans+") "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        if isinstance(f1,FORALL1) :
            f1.var.extend(self.var)
            return f1
        return FORALL1(self.var,f1)

    def PropQuant(self) : 
        (lres,lres1,lres2) = self.fils[0].PropQuant()
        L=self.var[:]
        lres1.extend(L[:])
        lres2.extend(L)
        return (lres,lres1,lres2)
    
    def write(self, output) :
        print('ne devrait pas passer par ici ! FORALL1:write !')
        output.write("(forall (")
        for v in self.var :
            output.write("(")
            output.write(v)
            output.write(" Bool)")
        output.write(") ")
        self.fils[0].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return 0
       
    def isPNF(self,niveau) :
        if not niveau :
            return False
        return self.fils[0].isPNF(niveau)
    
    def MaPl(self,D,Df,niv,justeX = False) :
        if (len(D) > 0) :
            print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (FORALL1) !")
        # if niv != -1 :
        #     print("Warning ! Mauvaise valeur pour init dans la mise a plat d'une formule (FORALL1) !")
        return FORALL1(self.var,self.fils[0].MaPl(D,Df,niv,justeX))
    
        # if (len(D) > 0)  :
        #     print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (FORALL1) !")
        # sf=self.fils[0]
        # if isinstance(sf,FORALL) or isinstance(sf,FORALL1) or isinstance(sf,EXISTS) or isinstance(sf,EXISTS) : 
        #     aux = self.fils[0].MaPl(D,Df)
        # else :
        #     aux = sf.MaPl(D,Df)
        #     aux = AuxMaPl(aux,D,Df)
        # return FORALL1(self.var,aux)

    def AjoutPropQuant(self,L) :
        return FORALL1(self.var,self.fils[0].AjoutPropQuant(L))
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        if (not stp == krip.start) :
            print("Warning ! FFlat2z3 - FORALL : pb d'arguments !")
        if len(self.var)==0 :
            self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        (PropB, PropBV, PropQ1, PropQ1FA) = VarQ
        fout.write("(forall (")
        for v in self.var :
            if v in PropQ1 : 
                fout.write("("+v+" (_ BitVec "+str(krip.TBVQ)+"))")
            else :
                print("Cas bizarre dans FFlat2z3 - FORALL1 !")
        fout.write(")")
        self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        fout.write(")")
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return EXISTS1(self.var,f)
        else :
            return FORALL1(self.var,f)

###############################################################################
#                          EXISTS
###############################################################################
class EXISTS :
    def __init__(self, varList, a) :
        self.fils=[a]
        self.var=varList

    def toString(self) :
        #print("Exist toString :")
        #print (self.fils)
        #print(self.varList)
        ans=""
        for v in self.var :
            ans=ans+"("+v+" Bool)"
        return "(exists ("+ans+") "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        if isinstance(f1,EXISTS) :
            f1.var.extend(self.var)
            return f1
        return EXISTS(self.var,f1)

    def PropQuant(self) :
        (lres,lres1,lres2) = self.fils[0].PropQuant()
        L=self.var[:]
        lres.extend(L)
        return (lres,lres1,lres2)
        
    def write(self, output) :
        output.write("(exists (")
        for v in self.var :
            output.write("(")
            output.write(v)
            output.write(" Bool)")
        output.write(") ")
        self.fils[0].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        if t==None :
            return self.fils[0].nbalt(0)
        elif t==1 :
            return 1+self.fils[0].nbalt(0)
        else :
            return self.fils[0].nbalt(0)

    def isPNF(self,niveau) :
        if not niveau :
            return False
        return self.fils[0].isPNF(niveau)

    def MaPl(self,D,Df,niv,justeX = False) :
        if (len(D) > 0) :
            print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (EXISTS) !")
        # if niv != -1 :
        #     print("Warning ! Mauvaise valeur pour init dans la mise a plat d'une formule (EXISTS) !")
        return EXISTS(self.var,self.fils[0].MaPl(D,Df,niv,justeX))
    
        # if (len(D) > 0) : 
        #     print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (EXISTS) !")
        # sf=self.fils[0]
        # if isinstance(sf,FORALL) or isinstance(sf,FORALL1) or isinstance(sf,EXISTS) or isinstance(sf,EXISTS) :
        #     aux = self.fils[0].MaPl(D,Df)
        # else :
        #     aux = sf.MaPl(D,Df)
        #     aux = AuxMaPl(aux,D,Df)
        # return EXISTS(self.var,aux)

    def AjoutPropQuant(self,L) :
        return EXISTS(self.var,self.fils[0].AjoutPropQuant(L))
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        if (not stp == krip.start) :
            print("Warning ! FFlat2z3 - FORALL : pb d'arguments !")
        if len(self.var)==0 :
            self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        (PropB, PropBV, PropQ1, PropQ1FA) = VarQ
        fout.write("(exists (")
        for v in self.var :
            if v in PropB :
                for s in krip.accessibles[stp] : 
                    fout.write("("+v+"_"+Kripkestring_of_state(s)+" Bool)")
            elif v in PropBV :
                for s in krip.accessibles[stp] : 
                    fout.write("("+v+"_"+Kripkestring_of_state(s)+" (_ BitVec "+str(krip.TBVU)+"))")
            else :
                print("Cas bizarre dans FFlat2z3 - EXISTS !")
        fout.write(")")
        if isinstance(self.fils[0],EXISTS) or isinstance(self.fils[0],EXISTS1) or isinstance(self.fils[0],FORALL) or isinstance(self.fils[0],FORALL1) : 
            self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        else :   # C'est le dernier EXISTS !! [celui des modalites CTL]
            fout.write("\n (or ")
            for v in PropQ1FA : # on ecarte les cas ou leur valeurs sont audela du nb d'etats !
                fout.write(" (bvuge "+v+" "+krip.cNumero()+") ")
            fout.write("\n (and ")
            self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
            (Dp,Df) = D
            for fname in Df :
                Faux = Df[fname]
                pFaux = Dp[fname]
                Faux.AGFlat2z3(pFaux,VarQ,krip,stp,fout,D)
                fout.write("\n")
            #il faut integrer les AG(p<=> def) !
            # on ajoute les bornes sup pour les BV
            for v in PropBV :
                for s in krip.accessibles[stp]  :
                    fout.write(" (bvule "+v+"_"+Kripkestring_of_state(s)+" "+krip.cBUntil()+") ")
            for v in PropQ1 :
                    if v not in PropQ1FA : # on ne restreint que les Exists_1
                        fout.write(" (bvule "+v+" "+krip.cNumero()+") ")                
            fout.write("))")
        fout.write(")")
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return FORALL(self.var,f)
        else :
            return EXISTS(self.var,f)

###############################################################################
#                          EXISTS1
###############################################################################
# exists1 = chaque prop de varlist est vraie dans exactelment un etat du modele... 
class EXISTS1 :
    def __init__(self, varList, a) :
        self.fils=[a]
        self.var=varList

    def toString(self) :
        #print("Exist1 toString :")
        #print (self.fils)
        #print(self.varList)
        ans=""
        for v in self.var :
            ans=ans+"("+v+" Bool)"
        return "(exists1 ("+ans+") "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        if isinstance(f1,EXISTS1) :
            f1.var.extend(self.var)
            return f1
        return EXISTS1(self.var,f1)

    def PropQuant(self) : 
        (lres,lres1,lres2) = self.fils[0].PropQuant()
        L=self.var[:]
        lres1.extend(L)
        return (lres,lres1,lres2)
    
    def write(self, output) :
        print('ne devrait pas passer par ici ! EXISTS1:write !')
        output.write("(exists (")
        for v in self.var :
            output.write("(")
            output.write(v)
            output.write(" Bool)")
        output.write(") ")
        self.fils[0].write(output)
        output.write(")")

    def nbalt(self,t = None) :
        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return 0

    def isPNF(self,niveau) :
        if not niveau :
            return False
        return self.fils[0].isPNF(niveau)
    
    def MaPl(self,D,Df,niv,justeX = False) :
        if (len(D) > 0) :
            print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (EXISTS1) !")
        # if niv != -1 :
        #     print("Warning ! Mauvaise valeur pour init dans la mise a plat d'une formule (EXISTS1) !")
        return EXISTS1(self.var,self.fils[0].MaPl(D,Df,niv,justeX))
    
        # if (len(D) > 0) : 
        #     print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme Prenex (EXISTS1) !")
        # sf=self.fils[0]
        # if isinstance(sf,FORALL) or isinstance(sf,FORALL1) or isinstance(sf,EXISTS) or isinstance(sf,EXISTS) :
        #     aux = self.fils[0].MaPl(D,Df)
        # else :
        #     aux = sf.MaPl(D,Df)
        #     aux = AuxMaPl(aux,D,Df)
        # return EXISTS1(self.var,aux)

    def AjoutPropQuant(self,L) :
        return EXISTS1(self.var,self.fils[0].AjoutPropQuant(L))
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        if (not stp == krip.start) :
            print("Warning ! FFlat2z3 - FORALL : pb d'arguments !")
        if len(self.var)==0 :
            self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        (PropB, PropBV, PropQ1, PropQ1FA) = VarQ
        fout.write("(exists (")
        for v in self.var :
            if v in PropQ1 : 
                fout.write("("+v+" (_ BitVec "+str(krip.TBVQ)+"))")
            else :
                print("Cas bizarre dans FFlat2z3 - EXISTS1 !")
        fout.write(")")
        self.fils[0].FFlat2z3(VarQ,krip,stp,fout,D)
        fout.write(")")
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return FORALL1(self.var,f)
        else :
            return EXISTS1(self.var,f)

###############################################################################
#                          BOTTOM
###############################################################################
# vrai
class BOTTOM :
    def __init__(self) :
        self.fils=[]
    
    def toString(self) :
        return "false"

    def SimplifB(self) :
        return BOTTOM()

    def PropQuant(self) : 
        return ([],[],[])

    def write(self, output) :
        output.write("false")

    def nbalt(self,t = None) :
        return 0

    def isPNF(self,niveau) :
        return True
    
    def MaPl(self,D,Df,niv,justeX = False) : 
        return BOTTOM()

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write(" false ")
        
    def FNN(self,np) :
        if np :
            return TOP()
        else :
            return BOTTOM()

###############################################################################
#                          TOP
###############################################################################
class TOP :
    def __init__(self) :
        self.fils=[]
    
    def toString(self) :
        return "true"

    def SimplifB(self) :
        return TOP()

    def PropQuant(self) : 
        return ([],[],[])
    
    def write(self, output) :
        output.write("true")

    def nbalt(self,t = None) :
        return 0

    def isPNF(self,niveau) :
        return True
    
    def MaPl(self,D,Df,niv,justeX = False) : 
        return TOP()

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write(" true ")
        
    def FNN(self,np) :
        if np :
            return BOTTOM()
        else :
            return TOP()
###############################################################################
#                          EF
###############################################################################
class EF :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(EF "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f=self.fils[0].SimplifB()
        if isinstance(f,TOP) or isinstance(f,BOTTOM)  :
            return f
        else:
            return EF(f)

    def PropQuant(self) : 
        return self.fils[0].PropQuant()
    
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return self.fils[0].nbalt(t)

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False)

    def MaPl(self,D,Df,niv,justeX = False) :
        f1 = self.fils[0].MaPl(D,Df,1,False)
        if niv==-1 :
            return AuxMaPl(EF(f1),D,Df)
        elif niv==0 and not justeX : 
            return EF(f1)
        else : 
            aux=EF(f1)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        for st in krip.accessibles[s] :
            fout.write("(=> ")   # => vs <=>
            fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(0)+") ")
            f1.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(") ")
            for d in range(1,krip.BUntilD()) :
                fout.write("(=> ")    # => vs <=>
                fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(d)+") ")
                if len(krip.succ[st])==1 :
                    for stp in krip.succ[st] :
                        fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d-1)+") ")
                else :
                    fout.write("(or ")
                    for stp in krip.succ[st] :
                        fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d-1)+") ")
                        fout.write(" ")
                    fout.write(")")
                fout.write(")")
        fout.write("\n")
        
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
                
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return AG(f)
        else :
            return EF(f)

###############################################################################
#                          AF
###############################################################################       
class AF :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(AF "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f=self.fils[0].SimplifB()
        if isinstance(f,TOP) or isinstance(f,BOTTOM)  :
            return f
        else:
            return AF(f)

    def PropQuant(self) : 
        return self.fils[0].PropQuant()
    
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return self.fils[0].nbalt(t)

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False)

    def MaPl(self,D,Df,niv,justeX = False) :
        f1 = self.fils[0].MaPl(D,Df,1,False)
        if niv==-1 :
            return AuxMaPl(AF(f1),D,Df)
        elif niv==0 and not justeX : 
            return AF(f1)
        else : 
            aux=AF(f1)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        for st in krip.accessibles[s] :
            fout.write("(=> ")   # => vs <=>
            fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(0)+") ")
            f1.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(") ")
            for d in range(1,krip.BUntilD()) :
                fout.write("(=> ")     # => vs <=>
                fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(st)+" "+krip.cDistance(d)+") ")
                if len(krip.succ[st])==1 :
                    for stp in krip.succ[st] :
                        fout.write(" (= "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d-1)+") ")
                else :
                    fout.write(" (and ")
                    for stp in krip.succ[st] :
                        fout.write(" (bvult "+pFaux+"_"+Kripkestring_of_state(stp)+" "+krip.cDistance(d)+") ")
                        fout.write(" ")
                    fout.write(")")
                fout.write(")")
        fout.write("\n")
        
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return EG(f)
        else :
            return AF(f)
###############################################################################
#                          EG
###############################################################################
class EG :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(EG "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f=self.fils[0].SimplifB()
        if isinstance(f,TOP) or isinstance(f,BOTTOM)  :
            return f
        else:
            return EG(f)

    def PropQuant(self) : 
        return self.fils[0].PropQuant()
        
    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return self.fils[0].nbalt(t)


    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False)
    
    def MaPl(self,D,Df,niv,justeX = False) :
        f1 = self.fils[0].MaPl(D,Df,1,False)
        if niv==-1 :
            return AuxMaPl(EG(f1),D,Df)
        elif niv==0 and not justeX: 
            return EG(f1)
        else : 
            aux=EG(f1)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)

    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(=> ")     # => vs <=>  
            fp.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" (and ")
            f1.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" ")
            if len(krip.succ[st])==1 :
                for stp in krip.succ[st] : 
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
            else :
                fout.write("(or ")
                for stp in krip.succ[st] :
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
                    fout.write(" ")
                fout.write(")")
            fout.write("))")
        fout.write("\n")
        
    def FFlat2z3(self,PBool,PBT,VarQ,krip,stp) :
        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return AF(f)
        else :
            return EG(f)
            
###############################################################################
#                          AG
###############################################################################
class AG :
    def __init__(self, a) :
        self.fils=[a]

    def toString(self) :
        return "(AG "+self.fils[0].toString()+")"

    def SimplifB(self) :
        f=self.fils[0].SimplifB()
        if isinstance(f,TOP) or isinstance(f,BOTTOM)  :
            return f
        else:
            return AG(f)
        
    def PropQuant(self) : 
        return self.fils[0].PropQuant()

    def nbalt(self,t = None) :
#        print('Erreur : calcul du nb d alternations pour une formule QCTL !')
        return self.fils[0].nbalt(t)
        

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(False)

    def MaPl(self,D,Df,niv,justeX = False) :
        f1 = self.fils[0].MaPl(D,Df,1,justeX)
        if niv==-1 :
            return AuxMaPl(AG(f1),D,Df)
#        elif niv==0 and not justeX: # MODIF V1.1 -> supprimer le "and not justeX" 
        elif niv==0 : 
            return AG(f1)
        else : 
            aux=AG(f1)
            index=aux.toString()
            if index in D :
                prop = D[index]
            else :
                prop = getIdent()
                D[index]=prop
                Df[index]=aux
        return AP(prop)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def AGFlat2z3(self,pFaux,VarQ,krip,s,fout,D) :  # ecrit sur fout une liste (conj) de formules propo pour coder les def de AG(p<=> def)
        f1=self.fils[0]    # pour tout st on doit avoir: p_phi_st <-> \/_{st->st'} p_f1_st'
        fp=AP(pFaux)
        for st in krip.accessibles[s] :
            fout.write("(=> ")    # => vs <=>
            fp.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" (and ")
            f1.FFlat2z3(VarQ,krip,st,fout,D)
            fout.write(" ")
            if len(krip.succ[st])==1 :
                for stp in krip.succ[st] : 
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
            else :
                for stp in krip.succ[st] :
                    fp.FFlat2z3(VarQ,krip,stp,fout,D)
                    fout.write(" ")
            fout.write("))")
        fout.write("\n")
        
    def FFlat2z3(self,VarQ,krip,st,fout,D) :    #MODIF V1.1 -> renvoyer la conjonction sur tous les etats !
#        print("Warning ! FFlat2z3 ne devrait pas etre appele sur "+str(type(self)))
        f1=self.fils[0]
        if len(krip.accessibles[st])==1 :
                for stp in krip.accessibles[st] : 
                        f1.FFlat2z3(VarQ,krip,stp,fout,D)
        else :
            fout.write("(and ")
            for stp in krip.accessibles[st] :# on doit avoir:  /\_{st->*st'} p_f1_st' 
                f1.FFlat2z3(VarQ,krip,stp,fout,D) 
            fout.write(")")
            
    def FNN(self,np) :
        f=self.fils[0].FNN(np)
        if np :
            return EF(f)
        else :
            return AG(f)

###############################################################################
#                          EQ
###############################################################################    
# Equivalence
class EQ :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        return "(= "+self.fils[0].toString()+" "+self.fils[1].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        f2=self.fils[1].SimplifB()
        if isinstance(f1,TOP) :
          return f2
        elif isinstance(f2,TOP) :
          return f1
        elif isinstance(f1,BOTTOM) :
          return f2.FNN(True) # on fait la negation (et on la propage)
        elif isinstance(f2,BOTTOM) :
          return f1.FNN(True)
        else :
          return EQ(f1,f2)

    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)
        return (L,L1,L2)
      
    def write(self, output) :
      output.write("(= ")
      self.fils[0].write(output)
      output.write(" ")
      self.fils[1].write(output)
      output.write(")")

    def nbalt(self,t = None) :
        if t==None :
            return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))
        else :
            return max(self.fils[0].nbalt(0),self.fils[1].nbalt(1),self.fils[0].nbalt(1),self.fils[1].nbalt(0))

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(niveau) and self.fils[1].isPNF(niveau)

    def MaPl(self,D,Df,niv,justeX = False) :
        print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme NN (EQ) !")
        f1=self.fils[0].MaPl(D,Df,max(0,niv),justeX)
        f2=self.fils[1].MaPl(D,Df,max(0,niv),justeX)
        if niv==-1 : # on doit ajouter le bloc Exists
            return AuxMaPl(EQ(f1,f2),D,Df)
        return EQ(f1,f2)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout) :
        fout.write("(= "+self.fils[0].FFlat2z3(VarQ,krip,stp,fout)+" "+ self.fils[1].FFlat2z3(VarQ,krip,stp,fout)  +")")
        
    def FNN(self,np) :
        f1=self.fils[0].FNN(False)
        f1n=self.fils[0].FNN(True)
        f2=self.fils[1].FNN(False)
        f2n=self.fils[1].FNN(True)
        if np :
            return OR(AND(f1n,f2),AND(f1,f2n))
        else :
            return OR(AND(f1,f2),AND(f1n,f2n))
            
###############################################################################
#                          IMPL
###############################################################################        
class IMPL :
    def __init__(self, a1, a2) :
        self.fils=[a1, a2]

    def toString(self) :
        return "(=> "+self.fils[0].toString()+" "+self.fils[1].toString()+")"

    def SimplifB(self) :
        f1=self.fils[0].SimplifB()
        f2=self.fils[1].SimplifB()
        if isinstance(f1,TOP) :
          return f2
        elif isinstance(f1,BOTTOM) :
          return TOP()
        else :
          return IMPL(f1,f2)

    def PropQuant(self) : 
        (L,L1,L2)=self.fils[0].PropQuant()
        (M,M1,M2)=self.fils[1].PropQuant()
        L.extend(M)
        L1.extend(M1)
        L2.extend(M2)
        return (L,L1,L2)
      
    def write(self, output) :
      output.write("(=> ")
      self.fils[0].write(output)
      output.write(" ")
      self.fils[1].write(output)
      output.write(")")

    def nbalt(self,t = None) :
        if t==None :
            return max(self.fils[0].nbalt(t),self.fils[1].nbalt(t))
        elif t==0 :
            return max(self.fils[0].nbalt(1),self.fils[1].nbalt(0))
        else :
            return max(self.fils[0].nbalt(0),self.fils[1].nbalt(1))
        return 0

    def isPNF(self,niveau) :
        return self.fils[0].isPNF(niveau) and self.fils[1].isPNF(niveau)

    def MaPl(self,D,Df,niv,justeX = False) : 
        print("Warning ! Mise-a-plat d'une formule qui n'est pas sous forme NN (IMPL) !")
        f1=self.fils[0].MaPl(D,Df,max(0,niv),justeX)
        f2=self.fils[1].MaPl(D,Df,max(0,niv),justeX)
        if niv==-1 :
            return AuxMaPl(IMPL(f1,f2),D,Df)
        return IMPL(f1,f2)

    def AjoutPropQuant(self,L) :
        return EXISTS(L,self)
    
    def FFlat2z3(self,VarQ,krip,stp,fout,D) :
        fout.write("(=> "+self.fils[0].FFlat2z3(VarQ,krip,stp,fout)+" "+ self.fils[1].FFlat2z3(VarQ,krip,stp,fout)  +")")
        
    def FNN(self,np) :
        f1=self.fils[0].FNN(not np)
        f2=self.fils[1].FNN(np)
        if np :
            return AND(f1,f2)
        else :
            return OR(f1,f2)

########################################################################

global countAtLeast
countAtLeast=0

def incAtLeast() :
    global countAtLeast
    countAtLeast+=1
    return countAtLeast

# Exprime le fait qu'au moins bound successeurs satisfont formule.
def getXAtLeast(bound, formule) :
    if bound == 0 :
        return TOP()
    if bound == 1 :
        return EX(formule)
    global countAtLeast
    tmp=countAtLeast
    
    # La i-eme variable quantifiee qu'on utilisera.
    def getIdent(i) :
        return "atLeast_"+str(i+tmp)
    
    # Un bloc de i quantificateurs existentiels devant f.
    def putExists(i, f) :
        if i == 0 :
            return f
        var=[]
        for j in range(i) :
            var.append(getIdent(j+1))
        return EXISTS(var, f)
    
    log=(bound-1).bit_length()
    for j in range(log) :
        incAtLeast()
        
    # Une formule qui exprime qu'on se trouve dans le i-eme successeur satisfaisant formule.
    def getConj(i) :
        ans=[formule]
        # Ecriture de i en base 2.
        for j in range(log) :
            if i%2 == 0 :
                ans.append(NOT(AP(getIdent(j+1))))
            else :
                ans.append(AP(getIdent(j+1)))
            i=i//2
        return SEQAND(ans)
    
    ans=[]
    for i in range(bound) :
        ans.append(EX(getConj(i)))
    return putExists(log, SEQAND(ans))

# Exprime le fait qu'au moins bound accessibles satisfont formule.
def getFAtLeast(bound, formule) :
    if bound == 0 :
        return TOP()
    if bound == 1 :
        return EF(formule)
    global countAtLeast
    tmp=countAtLeast

    # La i-eme variable quantifiee qu'on utilisera.
    def getIdent(i) :
        return "atLeast_"+str(i+tmp)

    # Un blocde i quantificateurs existentiels devant f.
    def putExists(i, f) :
        if i == 0 :
            return f
        var=[]
        for j in range(i) :
            var.append(getIdent(j+1))
        return EXISTS(var, f)
        
    log=(bound-1).bit_length()
    for j in range(log) :
        incAtLeast()

    # Une formule qui exprime que l'on se trouve dans le i-eme accessible satisfaisant formule.
    def getConj(i) :
        ans=[formule]
        # Ecriture de i en base 2.
        for j in range(log) :
            if i%2 == 0 :
                ans.append(NOT(AP(getIdent(j+1))))
            else :
                ans.append(AP(getIdent(j+1)))
            i=i//2
        return SEQAND(ans)

    ans=[]
    for i in range(bound) :
        ans.append(EF(getConj(i)))

    return putExists(log, SEQAND(ans))

# Exprime le fait qu'exactement n successeurs satisfont formule.
def getXExactly(n, formule) :
    return AND(getXAtLeast(n, formule), NOT(getXAtLeast(n+1, formule)))

# Exprime le fait qu'exactement n accessibles satisfont formule.
def getFExactly(n, formule) :
    return AND(getFAtLeast(n, formule), NOT(getFAtLeast(n+1, formule)))

# Une fonction qui calcule par memoisation le nombre de noeuds de l'AST d'une formule.

tmp={}

def findLength(formule) :
    global tmp
    if formule in tmp :
        return tmp[formule]
    ans=1
    for fils in formule.fils :
        ans+=findLength(fils)
    tmp[formule]=ans
    return ans

# fct pour la mise a plat: on renvoie la formule:
# Exists[p1... pk]. f ^  /\_i AG (pi <=> Fi])   avec D[Fi]=pi

# D associe une prop a une string d'une formule
# Df associe une formule  a une string (de cette formule)
def AuxMaPl(f,D,Df) :
    if len(D)==0 :
        return f
    Lres = [f]
    Lprop = [ ]
    for fi in D :
        Lres.append(AG(EQ(AP(D[fi]),Df[fi])))  # pb: ici on veut la formule fi [pas la (string) fi] !!
        Lprop.append(D[fi])
    Fres = EXISTS(Lprop,SEQAND(Lres))
    return Fres
