from formule import *
from nnf import toNNF
from structure import Krip
import os
import sys
import time

cz3='z3'
#cz3='./z3'

# On veut ici, a partir d'une formule de QCTL et d'une structure de Kripke pointee,
# construire une formule de QBF qui est vraie si, et seulement si, la structure satisfait la formule.
# On a quatre differentes methodes pour le faire :

#    La premiere exprime les AF et EG a partir de EU et AU.
#    Les EU et AU sont ensuite vus comme des points fixes, et sont transformes en formules QCTL qui dependent de AG.

#    La deuxieme exprime les les AF et EG a partir de EU et AU.
#    Un [E f1 U f2] est ensuite transforme en [f2 | (f1 & EX (f2 | f1 & EX(... ], autant de fois qu'il y a d'etats accessibles.
#    De meme pour AU.

#    La troisieme mets d'abord la formule QCTL sous forme normale negative (ce qui nous force a introduire les AW et EW).
#    Les AF et EG sont ensuite exprimes a l'aide de AU, EU, AW et EW (en prenant soin de toujours avoir une forme normale negative).
#    Les EU, AU, EW, AW sont ensuite transformes en une formule QBF a l'aide d'un parcours intelligent de la structure
#    (un parcours en profondeur qui traite intelligemment les boucles).
#    Plus de details sont donnes dans le code.

#    La dernire fait la mme chose que la premire, mais sans introduire d'alternation de quantificateur.

#    Les quatre methodes expriment les EF (resp. AG) comme une disjonction (resp. conjonction) sur les etats accessibles.

#    En pratique la deuxieme methode se revele bien moins efficace que les deux autres.
#    Il existe des cas ou la premiere methode est plus efficace que la troisieme, et vice-versa.

# On essaiera de construire une formule la plus optimisee possible. Par exemple, on retirera
# les connecteurs TOP des conjonctions et les connecteurs BOTTOM des disjonctions. Si la formule
# a transformer est dans CTL, on aura directement trouve vrai ou faux. Sinon, la formule QBF produite
# ne contiendra aucun connecteur TOP ou BOTTOM : les sous-formules atomiques ne seront que des variables quantifiees.



# Dans les methodes 1 et 4, on cree des nouvelles quantifications pour exprimer les points fixes.
# Il faut donc prendre soin de choisir des identifiants inutilises.
# count et getIdent nous permettent de ne pas creer deux quantifications sur le meme identifiant.
# Il faut supposer que la formule QCTL de base ne contienne aucune proposition atomique de la forme "quant_n", ou n est un entier.
count=0

def getIdent() :
    global count
    count+=1
    return "quant_"+str(count)

# Dans les methodes qui suivent, find est une fonction qui prends un etat s (pas forcement l'etat initial) et une formule QCTL f,
# et qui renvoie une formule QBF vraie si, et seulement si, la structure de Kripke avec s comme etat initial satisfait f.
# Le code de find est donne plus tard.

def treatEU1(formule, find, state) :
    f1=formule.fils[0]
    f2=formule.fils[1]

    # Renvoie une formule QCTL qui se sert de AG et est equivalente a E f U g.
    # Cette formule contiendra deux quantifications dont ident1 et ident2 seront les variables.
    def unt(f, g, ident1, ident2) :
        
        # La fonction dont E f U g est le + petit point fixe.
        def funFix(h) :
            return OR(g, AND(f, EX(h)))
        
        # + petit point fixe de funFix
        return EXISTS([ident1], AND(AP(ident1), AND(AG(EQ(AP(ident1), funFix(AP(ident1)))), FORALL([ident2], IMPL(AG(EQ(AP(ident2), funFix(AP(ident2)))), AG(IMPL(AP(ident1), AP(ident2))))))))
    
    ident1=getIdent()
    ident2=getIdent()
    
    def estConnecteurTemporel(formule) :
        return not(isinstance(formule, AP) or isinstance(formule, NOT) or isinstance(formule, AND) or isinstance(formule, OR) or isinstance(formule, SEQAND) or isinstance(formule, SEQOR) or isinstance(formule, TOP) or isinstance(formule, BOTTOM) or isinstance(formule, IMPL) or isinstance(formule, EQ))
    
    def aConnecteurTemporel(formule) :
        for f in formule.fils :
            if estConnecteurTemporel(f) or aConnecteurTemporel(f) :
                return True
        return False
    
    if aConnecteurTemporel(formule) :
        return find(unt(f1, f2, ident1, ident2), state, [])
    else :
        ident3=getIdent()
        ident4=getIdent()
    
        # Permet de n'exprimer le until que sur ident1 et ident2, et non deux formules pouvant etre tres grandes.
        finalForm=EXISTS([ident1, ident2], AND(unt(AP(ident1), AP(ident2), ident3, ident4), AG(AND(IMPL(AP(ident1), f1), IMPL(AP(ident2), f2)))))

    return find(finalForm, state, [])

def treatEU4(formule, find, state) :
    # L'ide est la mme que la mthode 1.
    # On exprime cette fois-ci E f U g comme l'intersection de tous les points fixes.
    # On n'a alors besoin que d'un seul quantificateur.
    f1=formule.fils[0]
    f2=formule.fils[1]

    def unt(f, g, ident) :
        
        # La fonction dont E f U g est le + petit point fixe.
        def funFix(h) :
            return OR(g, AND(f, EX(h)))
        
        # pour tout ident, si ident est un point fixe, alors ident.
        return FORALL([ident], IMPL(AG(EQ(AP(ident), funFix(AP(ident)))), AP(ident)))
    
    ident=getIdent()
    
    # Permet de n'exprimer le until que sur ident1 et ident2, et non deux formules pouvant etre tres grandes.
    #finalForm=EXISTS([ident1, ident2], AND(unt(AP(ident1), AP(ident2), ident3, ident4), AG(AND(IMPL(AP(ident1), f1), IMPL(AP(ident2), f2)))))
    
    # Cette fois, si on veut simplifier le until en introduisant une quantification,
    # on va aussi introduire une alternation de quantificateurs, ce qui n'est peut-tre pas souhait.
    
    finalForm=unt(f1, f2, ident)
    
    return find(finalForm, state, [])

# EW4 : 26/03/2019
def treatEW4(formule, find, state) :
    # Identique a EU4 mais avec un + gd pt fixe ! 
    f1=formule.fils[0]
    f2=formule.fils[1]
    def unt(f, g, ident) :
        # La fonction dont E f W g est le + grand point fixe.
        def funFix(h) :
            return OR(g, AND(f, EX(h)))
        
        # pour tout ident, si ident est un point fixe, alors ident.
        return EXIST([ident], AND(AG(IMPL(AP(ident), funFix(AP(ident)))), AP(ident))  )
    
    ident=getIdent()
    finalForm=unt(f1, f2, ident)
    return find(finalForm, state, [])

def treatEU2(formule, find, state, krip) :
    f1=formule.fils[0]
    f2=formule.fils[1]
    def untilLen (i) :
        if i==0 :
            return f2
        return OR(f2, AND(f1, EX(untilLen(i-1))))
    ans=untilLen(len(krip.accessibles[state]))
    return find(ans, state, [])


def treatAU1(formule, find, state) :
    f1=formule.fils[0]
    f2=formule.fils[1]

    # Renvoie une formule QCTL qui se sert de AG et est equivalente a A f U g.
    # Cette formule contiendra deux quantifications dont ident1 et ident2 seront les variables.
    def unt(f, g, ident1, ident2) :

        # La fonction dont E f U g est le + petit point fixe.
        def funFix(h) :
            return OR(g, AND(f, AX(h)))
        
        # + petit point fixe de funFix
        return EXISTS([ident1], AND(AP(ident1), AND(AG(EQ(AP(ident1), funFix(AP(ident1)))), FORALL([ident2], IMPL(AG(EQ(AP(ident2), funFix(AP(ident2)))), AG(IMPL(AP(ident1), AP(ident2))))))))
    
    ident1=getIdent()
    ident2=getIdent()
    
    def estConnecteurTemporel(formule) :
        return not(isinstance(formule, AP) or isinstance(formule, NOT) or isinstance(formule, AND) or isinstance(formule, OR) or isinstance(formule, SEQAND) or isinstance(formule, SEQOR) or isinstance(formule, TOP) or isinstance(formule, BOTTOM) or isinstance(formule, IMPL) or isinstance(formule, EQ))
    
    def aConnecteurTemporel(formule) :
        for f in formule.fils :
            if estConnecteurTemporel(f) or aConnecteurTemporel(f) :
                return True
        return False
    
    if aConnecteurTemporel(formule) :
        return find(unt(f1, f2, ident1, ident2), state, [])
    else :
        ident3=getIdent()
        ident4=getIdent()
    
        # Permet de n'exprimer le until que sur ident1 et ident2, et non deux formules pouvant etre tres grandes.
        finalForm=EXISTS([ident1, ident2], AND(unt(AP(ident1), AP(ident2), ident3, ident4), AG(AND(IMPL(AP(ident1), f1), IMPL(AP(ident2), f2)))))
    return find(finalForm, state, [])

def treatAU4(formule, find, state) :
    # L'ide est la mme que la mthode 1.
    # On exprime cette fois-ci A f U g comme l'intersection de tous les points fixes.
    # On n'a alors besoin que d'un seul quantificateur.
    f1=formule.fils[0]
    f2=formule.fils[1]

    def unt(f, g, ident) :
        
        # La fonction dont A f U g est le + petit point fixe.
        def funFix(h) :
            return OR(g, AND(f, AX(h)))
        
        # pour tout ident, si ident est un point fixe, alors ident.
        return FORALL([ident], IMPL(AG(EQ(AP(ident), funFix(AP(ident)))), AP(ident)))
    
    ident=getIdent()
    
    # Permet de n'exprimer le until que sur ident1 et ident2, et non deux formules pouvant etre tres grandes.
    #finalForm=EXISTS([ident1, ident2], AND(unt(AP(ident1), AP(ident2), ident3, ident4), AG(AND(IMPL(AP(ident1), f1), IMPL(AP(ident2), f2)))))
    
    # Cette fois, si on veut simplifier le until en introduisant une quantification,
    # on va aussi introduire une alternation de quantificateurs, ce qui n'est peut-tre pas souhait.
    
    finalForm=unt(f1, f2, ident)
    
    return find(finalForm, state, [])

# AW4 : 26/03/2019
def treatAW4(formule, find, state) :
    # Identique a EU4 mais avec un + gd pt fixe ! 
    f1=formule.fils[0]
    f2=formule.fils[1]
    def unt(f, g, ident) :
        # La fonction dont E f W g est le + grand point fixe.
        def funFix(h) :
            return OR(g, AND(f, AX(h)))
        
        # pour tout ident, si ident est un point fixe, alors ident.
        return EXIST([ident], AND(AG(IMPL(AP(ident), funFix(AP(ident)))), AP(ident))  )
    
    ident=getIdent()
    finalForm=unt(f1, f2, ident)
    return find(finalForm, state, [])


def treatAU2(formule, find, state, krip) :
    f1=formule.fils[0]
    f2=formule.fils[1]
    def untilLen (i) :
        if i==0 :
            return f2
        return OR(f2, AND(f1, AX(untilLen(i-1))))
    return find(untilLen(len(krip.accessibles[state])), state, [])

# Les 4 fonctions qui suivent presentent la troisieme methode.
# L'idee est toujours la meme. On a une fonction recursive
# qu'on appelle sur l'etat courant. Elle marque cet etat,
# explore tous ses successeurs non marques, puis ENLEVE LA MARQUE
# de l'etat courant. Au lieu de faire un parcours en profondeur classique sur les ACCESSIBLES,
# on parcourt ainsi tous les CHEMINS (sans boucle) depuis l'etat courant. Pour des formules de la forme
# A/E f1 U/W f2, les etats marques representent alors un chemin de l'etat initial vers
# l'etat courant, dont tous les predecesseurs verifient f1 et pas f2.
# Si un successeur de l'etat courant est marque, on est tombe sur une boucle.
# Selon la formule qu'on regarde, la presence d'une boucle peut rendre plus simple la transformation vers QBF.

def treatEU3(formule, krip, find, state) :
    f1=formule.fils[0]
    f2=formule.fils[1]

    # On suppose que, dans le chemin qu'on examine, tous les etats marques verifient f1 et pas f2.
    marqued=set([])
    
    # Avec la supposition exprimee par marqued sur des predecesseurs de s,
    # On construit une formule QBF vraie si, et seulement si, s satisfait E f1 U f2.
    def xPlore(s) :
        marqued.add(s)
        ans1=find(f2, s, [])
        if isinstance(ans1, TOP) : # s satisfait f2
            marqued.remove(s)
            return TOP()
        ans2=find(f1, s, [])
        if isinstance(ans2, BOTTOM) :
            marqued.remove(s)
            return ans1
        
        bad1=isinstance(ans1, BOTTOM)
        good2=isinstance(ans2, TOP)
        # si bad1 est vrai,  alors il faut juste que s satisfasse f1 ET qu'un de ses sucesseurs satisfasse E f1 U f2
        # si good2 est vrai, alors il faut juste que s satisfasse f2 OU qu'un de ses sucesseurs satisfasse E f1 U f2
        
        ans3=[] # Disjonction sur les successeurs de s.
        for succ in krip.succ[s] :
            if not (succ in marqued) :
                f=xPlore(succ)
                    
                if isinstance(f, TOP) : # Un successeur satisfait E f1 U f2
                    marqued.remove(s)
                    if good2 : # On a trouve notre chemin.
                        return TOP()
                    if bad1 :
                        return ans2
                    return OR(ans1, ans2)
                
                if not isinstance(f, BOTTOM) : # Inutile d'ajouter faux a une disjonction.
                    ans3.append(f)

            # On ne fait rien si on trouve une boucle pour un EU.
                    
        marqued.remove(s)
        length=len(ans3)
        
        if length == 0 : # La disjonction vide est fausse.
            # Aucun successeur ne satisfait E f1 U f2
            # Il est donc necessaire et suffisant que s satisfasse f2
            return ans1

        # On retourne ans1 | (ans2 & disjonction(ans3)).
        if length == 1 :
            ans3=ans3[0]
        else :
            ans3=SEQOR(ans3)
        if good2 :
            ans2=ans3
        else :
            ans2=AND(ans2, ans3)
        if bad1 :
            return ans2
        return OR(ans1, ans2)
    return xPlore(state)

# L'idee est la meme que ci-dessus pour les trois fonctions qui suivent.

def treatAU3(formule, krip, find, state) :
    f1=formule.fils[0]
    f2=formule.fils[1]

    # On suppose que, dans le chemin qu'on examine, tous les etats marques verifient f1 et pas f2.
    marqued=set([])

    # Avec la supposition exprimee par marqued sur des predecesseurs de s,
    # On construit une formule QBF vraie si, et seulement si, s satisfait A f1 U f2.
    def xPlore(s) :
        marqued.add(s)
        ans1=find(f2, s, [])
        if isinstance(ans1, TOP) : # s satisfait f2
            marqued.remove(s)
            return TOP()
        ans2=find(f1, s, [])
        if isinstance(ans2, BOTTOM) :
            marqued.remove(s)
            return ans1

        # si bad1 est vrai,  alors il faut juste que s satisfasse f1 ET que tous ses sucesseurs satisfassent A f1 U f2
        # si good2 est vrai, alors il faut juste que s satisfasse f2 OU que tous ses sucesseurs satisfassent A f1 U f2
        bad1=isinstance(ans1, BOTTOM)
        good2=isinstance(ans2, TOP)

        ans3=[] # Conjonction sur les successeurs de s.
        for succ in krip.succ[s] :
            if not (succ in marqued) :
                f=xPlore(succ)

                # Un successeur ne satisfait pas A f1 U f2.
                # Il est donc necessaire et suffisant que s satisfasse f2
                if isinstance(f, BOTTOM) :
                    marqued.remove(s)
                    return ans1

                # Inutile d'ajouter vrai a une conjonction.
                if not isinstance(f, TOP) :
                    ans3.append(f)

            # Dans ce cas, on a trouve une boucle ou f1 est tout le temps vrai, et f2 seulement peu etre pour s.
            # Comme on est dans un AU, il est necessaire et suffisant que f satisfasse f2.
            else :
                marqued.remove(s)
                return ans1

        
        marqued.remove(s)
        length=len(ans3)
        
        if length == 0 : # La conjonction vide est vraie. Tous les successeurs satisfont A f1 U f2.
            if good2 :
                return TOP()
            if bad1 :
                return ans2
            return OR(ans1, ans2)

        # On retourne ans1 | (ans2 & conjonction(ans3))
        if length == 1 :
            if good2 :
                ans2=ans3[0]
            else :
                ans2=AND(ans2, ans3[0])
        else :
            if not good2 :
                ans3.append(ans2)
            ans2=SEQAND(ans3)
        if bad1 :
            return ans2
        return OR(ans1, ans2)
    return xPlore(state)

def treatEW3(formule, krip, find, state) :
    f1=formule.fils[0]
    f2=formule.fils[1]

    # On suppose que, dans le chemin qu'on examine, tous les etats marques verifient f1 et pas f2.
    marqued=set([])

    # Avec la supposition exprimee par marqued sur des predecesseurs de s,
    # On construit une formule QBF vraie si, et seulement si, s satisfait E f1 W f2.
    def xPlore(s) :
        marqued.add(s)
        ans1=find(f2, s, [])
        if isinstance(ans1, TOP) : # s satisfait f2
            marqued.remove(s)
            return TOP()
        ans2=find(f1, s, [])
        if isinstance(ans2, BOTTOM) :
            marqued.remove(s)
            return ans1

        # si bad1 est vrai,  alors il faut juste que s satisfasse f1 ET qu'un de ses sucesseurs satisfasse E f1 W f2
        # si good2 est vrai, alors il faut juste que s satisfasse f2 OU qu'un de ses sucesseurs satisfasse E f1 W f2
        bad1=isinstance(ans1, BOTTOM)
        good2=isinstance(ans2, TOP)

        ans3=[] # Disjonction sur les successeurs de s.
        for succ in krip.succ[s] :
            if not (succ in marqued) :
                f=xPlore(succ)

                # Un successeur satisfait E f1 W f2.
                if isinstance(f, TOP) :
                    marqued.remove(s)
                    if good2 :
                        return TOP()
                    if bad1 :
                        return ans2
                    return OR(ans1, ans2)

                # Il est inutile d'ajouter faux a une disjonction.
                if not isinstance(f, BOTTOM) :
                    ans3.append(f)

            # Si on tombe sur une boucle ou f1 est tout le temps vraie, le EW est satisfait.
            else :
                marqued.remove(s)
                if good2 :
                    return TOP()
                if bad1 :
                    return ans2
                return OR(ans1, ans2)

        # On va retourner ans1 | (ans2 & disjonction(ans3))
        marqued.remove(s)
        length=len(ans3)
        
        if length == 0 : # La disjonction vide est fausse.
            return ans1
        if length == 1 :
            ans3=ans3[0]
        else :
            ans3=SEQOR(ans3)
        if good2 :
            ans2=ans3
        else :
            ans2=AND(ans2, ans3)
        if bad1 :
            return ans2
        return OR(ans1, ans2)
    return xPlore(state)

def treatAW3(formule, krip, find, state) :
    f1=formule.fils[0]
    f2=formule.fils[1]

    # On suppose que, dans le chemin qu'on examine, tous les etats marques verifient f1 et pas f2.
    marqued=set([])

    # Avec la supposition exprimee par marqued sur des predecesseurs de s,
    # On construit une formule QBF vraie si, et seulement si, s satisfait A f1 W f2.
    def xPlore(s) :
        marqued.add(s)
        ans1=find(f2, s, [])
        if isinstance(ans1, TOP) : # s satisfait f2.
            marqued.remove(s)
            return TOP()
        ans2=find(f1, s, [])
        if isinstance(ans2, BOTTOM) :
            marqued.remove(s)
            return ans1

        # si bad1 est vrai,  alors il faut juste que s satisfasse f1 ET que tous ses sucesseurs satisfassent A f1 W f2
        # si good2 est vrai, alors il faut juste que s satisfasse f2 OU que tous ses sucesseurs satisfassent A f1 W f2
        bad1=isinstance(ans1, BOTTOM)
        good2=isinstance(ans2, TOP)

        ans3=[] # Conjonction sur les successeurs de s.
        for succ in krip.succ[s] :
            if not (succ in marqued) :
                f=xPlore(succ)

                # Un des successeurs ne satisfait pas A f1 W f2.
                # Il est donc necessaire et suffisant que s satisfasse f2.
                if isinstance(f, BOTTOM) :
                    marqued.remove(s)
                    return ans1

                # Inutile de rajouter vrai a une conjonction.
                if not isinstance(f, TOP) :
                    ans3.append(f)

            # Dans un AW, on ne fait rien si on rencontre une boucle.

        # On retourne ans1 | (ans2 & conjonction(ans3))
        marqued.remove(s)
        length=len(ans3)
        
        if length == 0 : # La conjonction vide est vraie.
            if good2 :
                return TOP()
            if bad1 :
                return ans2
            return OR(ans1, ans2)
        
        if length == 1 :
            if good2 :
                ans2=ans3[0]
            else :
                ans2=AND(ans2, ans3[0])
        else :
            if not good2 :
                ans3.append(ans2)
            ans2=SEQAND(ans3)
        if bad1 :
            return ans2
        return OR(ans1, ans2)
    ans=xPlore(state)
    return ans


# La grande methode qui construit une formule QBF vraie si, et seulement si,
# la structure krip avec state comme etat initial satisfait la formule QCTL formule.
# env est la liste des identificateurs des propositions atomiques quantifiees de notre sous-formule.
# Dans le cas ou une sous-formule doit etre calculee plusieurs fois, on memoise nos resultats.
# dico est alors un dictionnaire, qui a un couple (sous-formule, etat) associe (si on l'a deja calculee)
# la formule QBF vraie quand l'etat satisfait la sous-formule.
# mode est un entier (1, 2, 3, ou 4) qui indique quelle methode utiliser.

def transformation(krip, state, formule, env, dico, mode) :
    # On recherche d'abord la transformation de form dans dico au cas ou elle a ete memoisee.
    # Si on ne la trouve pas, on fait un appel recursif.
    # envPlus est une liste de variables quantifiees supplementaires a ajouter a env.
    # Par exemple, on change env quand on appelle find sur une sous-formule d'une quantification.
    def find(form, st, envPlus) :
        index=form.toString()
        if (index, st) in dico :
            return dico[index, st]
        ans=transformation(krip, st, form, env+envPlus, dico, mode)
        dico[index, st]=ans
        return ans
    
    if  isinstance(formule,AP) :
        if formule.ap in env : # quantifiee
            return QUANT(formule.ap+"_"+Krip.string_of_state(state))
        if formule.ap in krip.label[state] :
            return TOP()
        return BOTTOM()
    
    elif isinstance(formule, NOT) :
        formule2=find(formule.fils[0], state, [])
        if isinstance(formule2, TOP) :
            return BOTTOM()
        if isinstance(formule2, BOTTOM) :
            return TOP()
        if isinstance(formule2, NOT) :
            return formule2.fils[0]
        return NOT(formule2)
    
    elif isinstance(formule, OR) :
        fils1, fils2=find(formule.fils[0], state, []), find(formule.fils[1], state, [])
        if isinstance(fils1, TOP) or isinstance(fils2, TOP) :
            return TOP()
        if isinstance(fils1, BOTTOM) :
            return fils2
        if isinstance(fils2, BOTTOM) :
            return fils1
        return OR(fils1, fils2)
    
    elif isinstance(formule, AND) :
        fils1, fils2=find(formule.fils[0], state, []), find(formule.fils[1], state, [])
        if isinstance(fils1, BOTTOM) or isinstance (fils2, BOTTOM) :
            return BOTTOM()
        if isinstance(fils1, TOP) :
            return fils2
        if isinstance(fils2, TOP) :
            return fils1
        return AND(fils1, fils2)

    # Une disjonction sur tous les successeurs.
    elif isinstance(formule, EX) :
        fList=[]
        formule2=formule.fils[0]
        for st in krip.succ[state] :
            f=find(formule2, st, [])
            if isinstance(f, TOP) :
                return TOP()
            if isinstance(f, BOTTOM) : # Inutile d'ajouter faux a une disjonction.
                continue
            fList.append(f)
        length=len(fList)
        if length == 0 : # La disjonction vide est fausse.
            return BOTTOM()
        if length == 1 :
            return fList[0]
        return SEQOR(fList)

    # Une conjonction sur tous les successeurs.
    elif isinstance(formule, AX) :
        fList=[]
        formule2=formule.fils[0]
        for st in krip.succ[state] :
            f=find(formule2, st, [])
            if isinstance(f, BOTTOM) :
                return BOTTOM()
            if isinstance(f, TOP) : # Inutile d'ajouter vrai a une conjonction.
                continue
            fList.append(f)
        length=len(fList)
        if length == 0 : # La conjonction vide est vraie.
            return TOP()
        if len == 1 :
            return fList[0]
        return SEQAND(fList)

    elif isinstance(formule, AG) : # Conjonction sur tous les accessibles.
        andList=[]
        for s in krip.accessibles[state] :
            f=find(formule.fils[0], s, [])
            if isinstance(f, BOTTOM) :
                return BOTTOM()
            if not isinstance(f, TOP) : # Inutile d'ajouter vrai a une conjonction.
                andList.append(f)
        length=len(andList)
        if length == 0 : # La conjonction vide est vraie.
            return TOP()
        if length == 1 :
            return andList[0]
        return SEQAND(andList)

    elif isinstance(formule, AF) :
        return find(AU(TOP(), formule.fils[0]), state, [])

    elif isinstance(formule, EG) :
        return find(EW(formule.fils[0], BOTTOM()), state, [])

    elif isinstance(formule, EF) : # Disjonction sur tous les accessibles.
        orList=[]
        for s in krip.accessibles[state] :
            f=find(formule.fils[0], s, [])
            if isinstance(f, TOP) :
                return TOP()
            if not isinstance(f, BOTTOM) : # Inutile d'ajouter faux a une disjonction.
                orList.append(f)
        length=len(orList)
        if length == 0 :
            return BOTTOM() # La disjonction vide est fausse.
        if length == 1 :
            return orList[0]
        return SEQOR(orList)

    elif isinstance(formule, EU) :
        if mode == 1 :
            return treatEU1(formule, find, state)
        elif mode == 2 :
            return treatEU2(formule, find, state, krip)
        elif mode == 3 :
            return treatEU3(formule, krip, find, state)
        else :
        	return treatEU4(formule, find, state)

    elif isinstance(formule, AU) :
        if mode == 1 :
            return treatAU1(formule, find, state)
        elif mode == 2 :
            return treatAU2(formule, find, state, krip)
        elif mode == 3 :
            return treatAU3(formule, krip, find, state)
        else :
        	return treatAU4(formule, find, state)

    elif isinstance(formule, EW) :
        if mode == 3 :
            return treatEW3(formule, krip, find, state)
        else :
            return find(NOT(AU(NOT(formule.fils[1]), NOT(formule.fils[0]))), state, [])

    elif isinstance(formule, AW) :
        if mode == 3 :
            return treatAW3(formule, krip, find, state)
        else :
            return find(NOT(EU(NOT(formule.fils[1]), NOT(formule.fils[0]))), state, [])

    elif isinstance(formule, IMPL) :
        f1=find(formule.fils[0], state, [])
        f2=find(formule.fils[1], state, [])
        if isinstance(f1, BOTTOM) or isinstance(f2, TOP) :
            return TOP()
        if isinstance(f1, TOP) and isinstance(f2, BOTTOM) :
            return BOTTOM()
        if isinstance(f1, TOP) :
            return f2
        if isinstance(f2, BOTTOM) :
            if mode == 3 :
                return toNNF(NOT(f1))
            return NOT(f1)
        return IMPL(f1, f2)

    elif isinstance(formule, EQ) :
        f1=find(formule.fils[0], state, [])
        f2=find(formule.fils[1], state, [])
        if (isinstance(f1, TOP) and isinstance(f2, TOP)) or (isinstance(f1, BOTTOM) and isinstance(f2, BOTTOM)) :
            return TOP()
        elif (isinstance(f1, TOP) and isinstance(f2, BOTTOM)) or (isinstance(f1, BOTTOM) and isinstance(f2, TOP)) :
            return BOTTOM()
        return EQ(f1, f2)

    elif isinstance(formule, SEQAND) :
        transList=[]
        for fils in formule.fils :
            f=find(fils, state, [])
            if isinstance(f, BOTTOM) :
                return BOTTOM()
            if not isinstance(f, TOP) : # Inutile de rajouter vrai a une conjonction.
                transList.append(f)
        length=len(transList)
        if length == 0 : # La conjonction vide est vraie.
            return TOP()
        if length == 1 :
            return transList[0]
        return SEQAND(transList)

    elif isinstance(formule, SEQOR) :
        transList=[]
        for fils in formule.fils :
            f=find(fils, state, [])
            if isinstance(f, TOP) :
                return TOP()
            if not isinstance(f, BOTTOM) : # Inutile de rajouter faux a une disjonction.
                transList.append(f)
        length=len(transList)
        if length == 0 : # La disjonction vide est fausse.
            return BOTTOM()
        if length == 1 :
            return transList[0]
        return SEQOR(transList)

    # On multiplie le nombre de variables quantifiees par le nombre d'etats accessibles.
    elif isinstance(formule, FORALL) :
        varList=[]
        fils=find(formule.fils[0], state, formule.var)
        if isinstance(fils, TOP) or isinstance(fils, BOTTOM) :
            return fils
        for var in formule.var :
            for st in krip.accessibles[state] :
                varList.append(var+"_"+Krip.string_of_state(st))
        return FORALL(varList, fils)
    
    elif isinstance(formule, FORALL1) :
        #return find(NOT(EXISTS1(formule.var,NOT(formule.fils[0]))), state, [])
        def conj(etat, var, krip, state) :
            ans=[]
            for accessible in krip.accessibles[state] :
                if accessible == etat :
                    ans.append(AP(var+"_"+Krip.string_of_state(accessible)))
                else :
                    ans.append(NOT(AP(var+"_"+Krip.string_of_state(accessible))))
            return SEQAND(ans)
        def disj(var,krip,state) :
            ans=[]    # Q? EST-CE QUE LA BONNE SEMANTIQUE NE SERAIT PAS: PR TT ETAT DE KTIKP ? (plutot que seulement les accessibles)  Non ?
            for accessible in krip.accessibles[state] :
                ans.append(conj(accessible, var, krip,state))
            return SEQOR(ans)
        varList=[]
        conjList=[ ]
        for var in formule.var :
            conjList.append(disj(var,krip,state))
            for st in krip.accessibles[state] :
                varList.append(var+"_"+Krip.string_of_state(st))
        aux2 = find(formule.fils[0], state, formule.var)
        aux1= SEQAND(conjList)
        auximpl = IMPL(aux1,aux2)
        f=FORALL(varList, auximpl )
        #print(f.toString())
        return f
    
    elif isinstance(formule, EXISTS) :
        varList=[]
        fils=find(formule.fils[0], state, formule.var)
        if isinstance(fils, TOP) or isinstance(fils, BOTTOM) :
            return fils
        for var in formule.var :
            for st in krip.accessibles[state] :  
                varList.append(var+"_"+Krip.string_of_state(st))
        return EXISTS(varList, fils)

    elif isinstance(formule, EXISTS1) :
        def conj(etat, var, krip, state) :
            ans=[]
            for accessible in krip.accessibles[state] :
                if accessible == etat :
                    ans.append(AP(var+"_"+Krip.string_of_state(accessible)))
                else :
                    ans.append(NOT(AP(var+"_"+Krip.string_of_state(accessible))))
            return SEQAND(ans)
        def disj(var,krip,state) :
            ans=[]    # Q? EST-CE QUE LA BONNE SEMANTIQUE NE SERAIT PAS: PR TT ETAT DE KTIKP ? (plutot que seulement les accessibles)  Non ?
            for accessible in krip.accessibles[state] :
                ans.append(conj(accessible, var, krip,state))
            return SEQOR(ans)
        varList=[]
        conjList=[find(formule.fils[0], state, formule.var)]
        for var in formule.var :
            conjList.append(disj(var,krip,state))
            for st in krip.accessibles[state] :
                varList.append(var+"_"+Krip.string_of_state(st))
        f=EXISTS(varList, SEQAND(conjList))
        #print(f.toString())
        return f

    elif isinstance(formule, TOP) or isinstance(formule, BOTTOM) :
        return formule
    
    raise "erreur"

#####
# TRANFORMATION SPECIALE POUR FORMULES PLATES
# F= Quantif (p ^ AG (p <=> def) ^ AG (p <=> def) ^ ...
def model_checking_PNF(krip, formule, fichname) :
    state=krip.start
    dico, env = { } , []
    Tstart=time.time()
    F = formule.FNN(False)
#    print("Nb d'alternations dans la formule initiale: ",F.nbalt())

    F=F.SimplifB()
    D,Df={ },{ }
    print("Negation Normal Form + Simplifications:")
    print(F.toString())
    F = F.MaPl(D,Df,0,True) 
    Lnewprop = [D[index] for index in D]
    if len(Lnewprop) > 0 : 
        F = F.AjoutPropQuant(Lnewprop) # ajoute un "dernier" bloc Exists q1...qk avant la formule aplatie
    print("Flattening (_simple_): ",F.toString())
    for index in D :
        print(D[index]," -> ",index)
    F=F.SimplifB()
    print("-----------------------------------------------------")
    print("Formula in fine:")
    print(F.toString())
    (VarQ,VarQone,VarQoneForall) = F.PropQuant()  # on ajoute une liste des Forall1 pour traitement special
     # on distingue dans VarQ, les var correspondant a un EU/AU/EF/AF
     # donc des BitVec...
    PropB, PropBV = [ ], [ ] # on va distinguer les vraies prop. booleennes (pour EW et AW) et les BitVec (pour EU et AU)
    for fname in Df :
        Faux = Df[fname]
        pFaux = D[fname]
        tofF = str(type(Faux))
        tofF=tofF[16:-2]
        if tofF in ['EU', 'AU', 'EF', 'AF'] :
            # ce sont des BitVec
            PropBV.append(pFaux)
        else :
            PropB.append(pFaux)
    # on complete avec les var quantif "normales" (exists/forall) non liees a des def:
    for v in VarQ :
        if v not in PropBV and v not in PropB :
            PropB.append(v)
            
    print("Quantified variables  : ")
    print("Quantified var. -> exists-1 et forall-1 :",str(VarQone))
    print("Quantified var. -> EU/AU/EF/AF :",str(PropBV))
    print("Quantified var. -> standard :",str(PropB))
# cette distinction sera utile pour le codage en Z3
    print("-----------------------------------------------------")
#    if os.path.isfile(fichname) :
#    	#        print("on supprime le vieux fichier out: ",fichname)
#    	if os.name != "nt" :
#    		os.system("rm "+fichname)
#    	else :
#    		os.system("del "+fichname)
    fout=open(fichname, 'w')
    fout.write("(assert ")
    F.FFlat2z3((PropB,PropBV,VarQone,VarQoneForall),krip,state,fout,(D,Df))
    fout.write(")\n")
    fout.write("(check-sat)\n")
    fout.close()
    Tend=time.time()
    print("Time for producing z3 code : "+str(Tend-Tstart))
    if os.name != "nt" :
    	os.system("time "+cz3+" "+fichname)
    else :
    	os.system(cz3+" "+fichname)
    
    # # on utilise D et Df pour coder les AG(p<=>def)
    # # NB:  Df[def_string] = def  et D[def_string]=p
    # Land = [ ] # la liste des termes de la confonction globale
    # for fname in Df :
    #     Faux = Df[fname]
    #     pFaux = D[fname]
    #     tofF = str(type(Faux))
    #     tofF=tofF[16:-2]
    #     if tofF in ['EU', 'AU', 'EW', 'AW', 'EX', 'AX', 'EF', 'AF', 'EG', 'AG'] :
    #         # OK on appelle...
    #         Land.extend(Faux.AGFlat2z3(pFaux,(PropB,PropBV,VarQone),krip,state))
    #     else :
    #         print("Warning : pb dans transformation5, un terme n'a pas la bonne forme...")
    #     print("affichage des termes de la conjonction:")
    # for f in Land :
    #     print(f.toString())
    # # ajout des contraintes liees aux "1-quantifications"
    # # ajout de la partie principale de phi (utilisant les defs)
    # # a faire ! 
    
    # return TOP()




##################################################################
def toQBF(krip, state, formule, mode) :
    print("formule a verifier :\n"+formule.toString())
    if mode == 3 :
        formule=toNNF(formule)
        print("mise sous NNF :\n"+formule.toString())

    return transformation(krip, state, formule, [], {}, mode)
